//-----------------------------------------------------------------------------
// name: VisualSine.cpp
// desc: hello sine wave, real-time
//
// author: Ge Wang (ge@ccrma.stanford.edu)
//   date: fall 2014
//   uses: RtAudio by Gary Scavone
//-----------------------------------------------------------------------------
#include "RtAudio/RtAudio.h"
#include "chuck.h"
#include <math.h>
#include <cmath> // added by Kadar
#include <stdlib.h>
#include <iostream>
using namespace std;

#ifdef __MACOSX_CORE__
#include <GLUT/glut.h>
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#endif

// FFT
#include "chuck_fft.h"

//-----------------------------------------------------------------------------
// function prototypes
//-----------------------------------------------------------------------------
void initGfx();
void idleFunc();
void displayFunc();
void reshapeFunc( GLsizei width, GLsizei height );
void keyboardFunc( unsigned char, int, int );
void mouseFunc( int button, int state, int x, int y );
void drawLissajous( SAMPLE * stereobuffer, int len, int channels);
double compute_log_spacing( int fft_size, double factor );
inline double map_log_spacing( double ratio, double power );
double compute_log_spacing( int fft_size, double power );


// our datetype
#define SAMPLE float
// sound buffers
#define SND_BUFFER_SIZE         1024
#define SND_FFT_SIZE            ( SND_BUFFER_SIZE * 2 )
// corresponding format for RtAudio
#define MY_FORMAT RTAUDIO_FLOAT32
// sample rate
#define MY_SRATE 44100
// number of channels
#define MY_CHANNELS_IN 1
#define MY_CHANNELS_OUT 2
#define GREATER(a,b) ( a > b ? a : b )
// for convenience
#define MY_PIE 3.14159265358979

// width and height
long g_width = 1024;
long g_height = 720;
// global buffer
SAMPLE * g_buffer = NULL;
SAMPLE * fft_buffer = NULL;
GLfloat g_log_positions[SND_FFT_SIZE / 2];
GLfloat g_fft_size = SND_FFT_SIZE;
long g_bufferSize;

// global variables
bool g_draw_dB = false;
ChucK * the_chuck = NULL;

// for waterfall
struct Pt2D { float x; float y; };
Pt2D ** g_spectrums = NULL;
GLuint g_depth = 69; // for john: 64
GLfloat g_z = 0.0f;
GLboolean g_z_set = FALSE;
GLfloat g_space = .12f; // for john: .1f
GLboolean g_downsample = FALSE;
GLint g_ds = 0; // downsample amount

// gain
GLfloat g_gain = 1.0f;
GLfloat g_time_scale = 1.0f;
GLfloat g_freq_scale = 1.0f;
GLfloat g_lissajous_scale = 0.45f;

// how much to see
GLint g_time_view = 1;
GLint g_freq_view = 2;

// for log scaling
GLdouble g_log_space = 0;
GLdouble g_log_factor = 1;

// the index associated with the waterfall
GLint g_wf = 0;
// starting a file
GLint g_starting = 0;

// array of booleans for waterfall
GLboolean * g_draw = NULL;
// plot waterfall
GLboolean g_wutrfall = TRUE;

// for time domain waterfall
SAMPLE ** g_waveforms = NULL;
GLfloat g_wf_delay_ratio = 1.0f / 3.0f;
GLuint g_wf_delay = (GLuint)(g_depth * g_wf_delay_ratio + .5f);
GLuint g_wf_index = 0;

// waveform line width
GLfloat g_wave_lineWidth = 1;
// spectrum line width
GLfloat g_freq_lineWidth = 1;

GLfloat g_window[SND_BUFFER_SIZE]; // DFT transform window

double x_vals[50];
double y_vals[50];
SAMPLE g_cur_buffer[SND_BUFFER_SIZE];  // current mono buffer (now playing), for lissajous
GLint g_delay = SND_BUFFER_SIZE/2;
SAMPLE g_back_buffer[SND_BUFFER_SIZE]; // for lissajous

// Kadar's globals
GLfloat height = 0;
GLfloat depth = -1;
GLfloat lineLength = 0;
GLfloat xStartLengthOffset = 0;
const double freqThreshold = 1.0387e-08;

float bottomLeftX = -5.0f;
float bottomLeftY = -3.0f;
float bottomLeftZ = 5.0f;

float topLeftX = -1.0f;
float topLeftY = 1.8f;
float topLeftZ = -88.0f;

float topRightX = topLeftX * -1;
float topRightY = topLeftY;
float topRightZ = topLeftZ;

float bottomRightX = bottomLeftX * -1;
float bottomRightY = bottomLeftY;
float bottomRightZ = bottomLeftZ;

float roadBumperFactor = 12.4;

struct roadLine {
    float x;
    float y;
    float z;
};

struct flyingObject {
    float size;
    float xRate;
    float zRate;
    float xRunning;
    float zRunning;
    float rotatingAngle;
};

vector<roadLine> g_roadlines;
deque<flyingObject> g_objects;

bool isTrip = false;
bool isNarrative = false;

float cubeColorX = 1.0;
float cubeColorY = 1.0;
float cubeColorZ = 1.0;

// ----
// name: about()
// desc: prints info about the program along with meaningful keystrokes
void about() {
    fprintf( stderr, "----------------------------------------------------\n" );
    fprintf( stderr, "SoundTrip (1.0)\n" );
    fprintf( stderr, "Kadar D. Qian\n" );
    fprintf( stderr, "----------------------------------------------------\n" );
    fprintf( stderr, "'a' - add more flying cubes\n" );
    fprintf( stderr, "*space bar* - enter trip mode. tap multiple times in trip mode to add more cubes and teapots\n" );
    fprintf( stderr, "'c' - enter chill mode\n" );
    fprintf( stderr, "'m' - use this key to visualize sound from the microphone. note: after the narrative finishes, you will need to hit 'm' to reactivate the mic input\n" );
    fprintf( stderr, "'n' - play the narrative\n" );
    fprintf( stderr, "'q' - exit SoundTrip\n" );
    fprintf( stderr, "'r' - remove a flying object from the screen\n" );
}

// Credit to function found on stack overflow: https://stackoverflow.com/questions/5289613/generate-random-float-between-two-floats
float getRandomFloat(float a, float b) {
    float random = ((float) rand()) / (float) RAND_MAX;
    float diff = b - a;
    float r = random * diff;
    return a + r;
}

void playNarrative() {
    isNarrative = true;
    the_chuck->compileFile("Narrative", "");
}

//-----------------------------------------------------------------------------
// name: callme()
// desc: audio callback
//-----------------------------------------------------------------------------
int callme( void * outputBuffer, void * inputBuffer, unsigned int numFrames,
            double streamTime, RtAudioStreamStatus status, void * data )
{
    // cast!
    SAMPLE * input = (SAMPLE *)inputBuffer;
    SAMPLE * output = (SAMPLE *)outputBuffer;

    if (isNarrative) {
        // compute chuck!
        the_chuck->run(input, output, numFrames);
        for( int i = 0; i < numFrames; i++ )
        {
            // visualize all channels averaged
            g_buffer[i] = 0;
            for( int j = 0; j < MY_CHANNELS_OUT; j++ )
            {
                // accumulate each channel
                g_buffer[i] += output[i*MY_CHANNELS_OUT+j];
            }
            // average by number of channels
            g_buffer[i] /= MY_CHANNELS_OUT;
        }
    } else {
        for( int i = 0; i < numFrames; i++ )
        {
            // copy the input to visualize only the left-most channel
            g_buffer[i] = input[i*MY_CHANNELS_IN];

            // mute output -- TODO will need to disable this once ChucK produces output, in order for you to hear it!
            for( int j = 0; j < MY_CHANNELS_OUT; j++ ) { output[i*MY_CHANNELS_OUT + j] = 0; }
        }
    }
    
    
    return 0;
}

// ----
// name: initChucK()
// desc: initializes ChucK
// ----

// use this to compile ChucK code
stringstream ss;

void initChucK() {
    // set up chuck
    the_chuck = new ChucK();
    
    char directory[1024];
    getcwd( directory, 1024);
    the_chuck->setParam( CHUCK_PARAM_WORKING_DIRECTORY, directory );
    
    // set sample rate and number of in/out channels on our chuck
    the_chuck->setParam(CHUCK_PARAM_SAMPLE_RATE, MY_SRATE);
    the_chuck->setParam(CHUCK_PARAM_OUTPUT_CHANNELS, MY_CHANNELS_OUT);
    the_chuck->setParam(CHUCK_PARAM_INPUT_CHANNELS, MY_CHANNELS_IN);
    
    // initialize our chuck
    if (!the_chuck->init()) {
        cerr << "Failed to initialize ChucK" << endl;
        exit(0);
    }
    
}

int tripCount = 0;

/* The ranges for randomness of flying objects */
float xRateHigh = .04f;
float xRateLow = .01f;
float zRateHigh = .05f;
float zRateLow = .2f;
float xStartLow = .1f;
float xStartHigh = .3f;

void initFlyingObjects(int quantity) {
    for (int i = 0; i < quantity; i++) {
        float size = getRandomFloat(.1, .25);
        
        float xRate = getRandomFloat(xRateLow, xRateHigh);
        float positiveOrNegative = getRandomFloat(0, 1);
        if (positiveOrNegative >= .5) {
            xRate *= -1;
        }
        
        float randomxStart = getRandomFloat(xStartLow, xStartHigh);
        float positiveOrNegative2 = getRandomFloat(0, 1);
        if (positiveOrNegative2 >= .5) {
            randomxStart *= -1;
        }
        
        float zRate = getRandomFloat(zRateLow, zRateHigh);
        struct flyingObject fo = {size, xRate, zRate, positiveOrNegative2, 0.0, 0.0};
        
        g_objects.push_back(fo);
    }
}

void removeFlyingObject() {
    if (g_objects.size() > 1) {
        g_objects.pop_front();
    }
}

void startTrip() {
    tripCount++;
    isTrip = true;
    
    xRateHigh *= 6;
    xRateLow *= -3;
    zRateHigh *= 3;
    zRateLow *= 1;
    xStartLow = .1f;
    xStartHigh = .3f;
    
    initFlyingObjects(10);
    roadBumperFactor = 420;
}

void stopTrip() {
    isTrip = false;
    
    for (int i = 0; i < tripCount; i++) {
        removeFlyingObject();
    }
    
    roadBumperFactor = 12.4;
    
    cubeColorX = 1.0;
    cubeColorY = 1.0;
    cubeColorZ = 1.0;
    
    xRateHigh = .04f;
    xRateLow = .01f;
    zRateHigh = .05f;
    zRateLow = .2f;
    xStartLow = .1f;
    xStartHigh = .3f;
}



//-----------------------------------------------------------------------------
// name: main()
// desc: entry point
//-----------------------------------------------------------------------------
int main( int argc, char ** argv )
{
    about();
    // instantiate RtAudio object
    RtAudio audio;
    // variables
    unsigned int bufferBytes = 0;
    // frame size
    unsigned int bufferFrames = 1024;
    
    // check for audio devices
    if( audio.getDeviceCount() < 1 )
    {
        // nopes
        cout << "no audio devices found!" << endl;
        exit( 1 );
    }
    
    // initialize GLUT
    glutInit( &argc, argv );
    // init gfx
    initGfx();
    
    // init blocks
    initFlyingObjects(5);
    
    // let RtAudio print messages to stderr.
    audio.showWarnings( true );
    
    // set input and output parameters
    RtAudio::StreamParameters iParams, oParams;
    iParams.deviceId = audio.getDefaultInputDevice();
    iParams.nChannels = MY_CHANNELS_IN;
    iParams.firstChannel = 0;
    oParams.deviceId = audio.getDefaultOutputDevice();
    oParams.nChannels = MY_CHANNELS_OUT;
    oParams.firstChannel = 0;
    
    // create stream options
    RtAudio::StreamOptions options;
    
    // go for it
    try {
        // open a stream
        audio.openStream( &oParams, &iParams, MY_FORMAT, MY_SRATE, &bufferFrames, &callme, (void *)&bufferBytes, &options );
    }
    catch( RtError& e )
    {
        // error!
        cout << e.getMessage() << endl;
        exit( 1 );
    }
    
    // compute
    bufferBytes = bufferFrames * GREATER(MY_CHANNELS_OUT, MY_CHANNELS_IN) * sizeof(SAMPLE);
    // allocate global buffer
    g_bufferSize = bufferFrames;
    g_buffer = new SAMPLE[g_bufferSize];
    memset( g_buffer, 0, sizeof(SAMPLE)*g_bufferSize );
    
    // inititalize ChucK!
    initChucK();
    
    // go for it
    try {
        the_chuck->start();
        
        // start stream
        audio.startStream();
        
        // let GLUT handle the current thread from here
        glutMainLoop();
        
        // stop the stream.
        audio.stopStream();
    }
    catch( RtError& e )
    {
        // print error message
        cout << e.getMessage() << endl;
        goto cleanup;
    }
    
cleanup:
    // close if open
    if( audio.isStreamOpen() )
        audio.closeStream();
    
    // done
    return 0;
}




//-----------------------------------------------------------------------------
// Name: reshapeFunc( )
// Desc: called when window size changes
//-----------------------------------------------------------------------------
void initGfx()
{
    // double buffer, use rgb color, enable depth buffer
    glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
    // initialize the window size
    glutInitWindowSize( g_width, g_height );
    // set the window postion
    glutInitWindowPosition( 100, 100 );
    // create the window
    glutCreateWindow( "SoundTrip" );
    
    // set the idle function - called when idle
    glutIdleFunc( idleFunc );
    // set the display function - called when redrawing
    glutDisplayFunc( displayFunc );
    // set the reshape function - called when client area changes
    glutReshapeFunc( reshapeFunc );
    // set the keyboard function - called on keyboard events
    glutKeyboardFunc( keyboardFunc );
    // set the mouse function - called on mouse stuff
    glutMouseFunc( mouseFunc );
    
    // set clear color
    glClearColor( 0, 0, 0, 1 );
    // enable color material
    glEnable( GL_COLOR_MATERIAL );
    // enable depth test
    glEnable( GL_DEPTH_TEST );
    
    // code taken from sndpeek to initialize g_spectums
    // initialize
    g_spectrums = new Pt2D *[g_depth];
    for( int i = 0; i < g_depth; i++ )
    {
        g_spectrums[i] = new Pt2D[SND_FFT_SIZE];
        memset( g_spectrums[i], 0, sizeof(Pt2D)*SND_FFT_SIZE );
    }
    g_draw = new GLboolean[g_depth];
    memset( g_draw, 0, sizeof(GLboolean)*g_depth );

    // compute log spacing
    g_log_space = compute_log_spacing( g_fft_size / 2, g_log_factor );
    
    g_waveforms = new SAMPLE *[g_wf_delay];
    for( int i = 0; i < g_wf_delay; i++ )
    {
        // allocate memory (stereo)
        g_waveforms[i] = new SAMPLE[g_bufferSize * 2];
        // zero it
        memset( g_waveforms[i], 0, g_bufferSize * 2 * sizeof(SAMPLE) );
    }
}




//-----------------------------------------------------------------------------
// Name: reshapeFunc( )
// Desc: called when window size changes
//-----------------------------------------------------------------------------
void reshapeFunc( GLsizei w, GLsizei h )
{
    // save the new window size
    g_width = w; g_height = h;
    // map the view port to the client area
    glViewport( 0, 0, w, h );
    // set the matrix mode to project
    glMatrixMode( GL_PROJECTION );
    // load the identity matrix
    glLoadIdentity( );
    // create the viewing frustum
    gluPerspective( 45.0, (GLfloat) w / (GLfloat) h, 1.0, 300.0 );
    // set the matrix mode to modelview
    glMatrixMode( GL_MODELVIEW );
    // load the identity matrix
    glLoadIdentity( );
    // position the view point
    gluLookAt( 0.0f, 0.0f, 10.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f );
}




//-----------------------------------------------------------------------------
// Name: keyboardFunc( )
// Desc: key event
//-----------------------------------------------------------------------------
void keyboardFunc( unsigned char key, int x, int y )
{
    switch( key )
    {
        case 'Q':
        case 'q':
            exit(1);
            break;
        case 'a': // add a flying object
            initFlyingObjects(1);
            break;
        case 'r': // remove one flying object
            removeFlyingObject();
            break;
        case 'd':
            g_draw_dB = !g_draw_dB;
            break;
        case 'c': // stop the trip
            stopTrip();
            break;
        case ' ': // start the trip
            startTrip();
            break;
        case 'm': // turn on microphone mode
            isNarrative = false;
            break;
        case 'n':
            playNarrative();
            break;
    }
    
    glutPostRedisplay( );
}




//-----------------------------------------------------------------------------
// Name: mouseFunc( )
// Desc: handles mouse stuff
//-----------------------------------------------------------------------------
void mouseFunc( int button, int state, int x, int y )
{
    if( button == GLUT_LEFT_BUTTON )
    {
        // when left mouse button is down
        if( state == GLUT_DOWN )
        {
        }
        else
        {
        }
    }
    else if ( button == GLUT_RIGHT_BUTTON )
    {
        // when right mouse button down
        if( state == GLUT_DOWN )
        {
        }
        else
        {
        }
    }
    else
    {
    }
    
    glutPostRedisplay( );
}




//-----------------------------------------------------------------------------
// Name: idleFunc( )
// Desc: callback from GLUT
//-----------------------------------------------------------------------------
void idleFunc( )
{
    // render the scene
    glutPostRedisplay( );
}

//-----------------------------------------------------------------------------
// Name: displayFunc( )
// Desc: callback function invoked to draw the client area
//-----------------------------------------------------------------------------
void displayFunc( )
{
    /* ==
     Time domain lines that denote the banks of the channel
     == */
    // local state
    static GLfloat zrot = 0.0f, c = 0.0f;
    // clear the color and depth buffers
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    // line width
    if (isTrip) {
        glLineWidth(16);
    } else {
        glLineWidth( 8 );
    }
    
    // Code from sndpeek for buffer
    SAMPLE *buffer[SND_FFT_SIZE];
    // get the latest (possibly preview) window
    memset( buffer, 0, SND_FFT_SIZE * sizeof(SAMPLE) );
    // copy currently playing audio into buffer
    memcpy( buffer, g_buffer, g_bufferSize * sizeof(SAMPLE) );
    
    // left channel bank
    glBegin( GL_LINE_STRIP );
    // color
    // increments
    float xDiff = (topLeftX - bottomLeftX) / g_bufferSize;
    float yDiff = (topLeftY - bottomLeftY) / g_bufferSize -.002;
    float zDiff = (topLeftZ - bottomLeftZ) / g_bufferSize;
    float xStart = bottomLeftX;
    float yStart = bottomLeftY;
    float zStart = bottomLeftZ;
    // loop over buffer
    for( int i = 0; i < g_bufferSize; i++ )
    {
        float fval = ((float)(g_bufferSize - i) / g_bufferSize);
        if (g_buffer[i] > .1) {
            fval = 1.2;
        }
        if (isTrip) {
            glColor3f(getRandomFloat(0.0, 1.0), getRandomFloat(0.0, 1.0), getRandomFloat(0.0, 1.0));
        } else {
            glColor3f( 0.855f * fval, 0.647f * fval, 0.125f * fval );
        }

        // plot
        glVertex3f( xStart, yStart + roadBumperFactor * g_buffer[i], zStart + i * (zDiff));
        // increment x
        xStart += xDiff;
        yStart += yDiff;
    }
    glEnd();

    xDiff = (topLeftX - bottomLeftX) / g_bufferSize;
    yDiff = (topLeftY - bottomLeftY) / g_bufferSize;
    zDiff = (topLeftZ - bottomLeftZ) / g_bufferSize;
    xStart = bottomLeftX;
    yStart = bottomLeftY;
    zStart = bottomLeftZ;
    
    float xEnd = bottomRightX;
    float yEnd = bottomRightY;
    float zEnd = 0;
    float xDiffEnd = (topRightX - bottomRightX) / g_bufferSize;
    float yDiffEnd = (topRightY - bottomRightY) / g_bufferSize;
    float zDiffEnd = (topRightZ - bottomRightZ) / g_bufferSize;
    
    // road
    glPushMatrix();
    if (isTrip) {
        glLineWidth(24);
    } else {
        glLineWidth( 12 );
    }
    for (int i = 0; i < g_bufferSize * 20; i++) {
        glBegin( GL_LINE_STRIP );
        float fval = ((float)((g_bufferSize * 2) - i) / (g_bufferSize * 1.5f));
        
        if (isTrip) {
            glColor3f(getRandomFloat(0.0, 1.0), getRandomFloat(0.0, 1.0), getRandomFloat(0.0, 1.0));
        } else {
            glColor3f( 0.4 * fval, .4 * fval, .4 * fval );
        }
        if (i % 2 == 0) {
            glVertex3f(xStart - .05, yStart, zStart);
            glVertex3f(xEnd + .05, yEnd, zStart);
        } else {
            glVertex3f(xEnd, yEnd, zStart);
            glVertex3f(xStart, yStart, zStart);
        }
        
        zStart += zDiff;
        glEnd();
    }
    glPopMatrix();
    
    glLineWidth(2.0);
    // middle line
    glBegin( GL_LINE_STRIP );
    // increments
    xDiff = (0 - 0) / g_bufferSize;
    yDiff = (0 - bottomRightY) / g_bufferSize;
    zDiff = (topRightZ - bottomRightZ) / g_bufferSize;
    xStart = 0;
    yStart = bottomRightY;
    zStart = bottomRightZ;
    
    // loop over buffer
    for( int i = 0; i < g_bufferSize; i++ )
    {
        float fval = ((float)(g_bufferSize - i) / g_bufferSize);
        if (g_buffer[i] > .1) {
            fval = 8.8;
        }
        glColor3f( 1.0 * fval, 1.0 * fval, 1.0 * fval );
        // plot
        glVertex3f( xStart + (roadBumperFactor / 2) * g_buffer[i], yStart , zStart + i * zDiff);
        // increment x
        xStart += xDiff;
        yStart += yDiff;
    }
    glEnd();
    
    if (isTrip) {
        glLineWidth(16);
    } else {
        glLineWidth( 8 );
    }
    // right channel bank
    glBegin( GL_LINE_STRIP );
    // increments
    xDiff = (topRightX - bottomRightX) / g_bufferSize;
    yDiff = (topRightY - bottomRightY) / g_bufferSize -.002;
    zDiff = (topRightZ - bottomRightZ) / g_bufferSize;
    xStart = bottomRightX;
    yStart = bottomRightY;
    zStart = bottomRightZ;
    // loop over buffer
    for( int i = 0; i < g_bufferSize; i++ )
    {
        float fval = ((float)(g_bufferSize - i) / g_bufferSize);
        if (g_buffer[i] > .1) {
            fval = 1.2;
        }
        if (isTrip) {
            glColor3f(getRandomFloat(0.0, 1.0), getRandomFloat(0.0, 1.0), getRandomFloat(0.0, 1.0));
        } else {
            glColor3f( 0.855f * fval, 0.647f * fval, 0.125f * fval );
        }
        // plot
        glVertex3f( xStart, yStart + roadBumperFactor * g_buffer[i] , zStart + i * zDiff);
        // increment x
        xStart += xDiff;
        yStart += yDiff;
    }
    glEnd();
    
    //     For the flying objects
    //     color
    for (int i = 0; i < g_objects.size(); i++) {
        glPushMatrix();
        if (isTrip) {
            cubeColorX = getRandomFloat(0.0, 1.0);
            cubeColorY = getRandomFloat(0.0, 1.0);
            cubeColorZ = getRandomFloat(0.0, 1.0);
        }
        glColor3f(cubeColorX, cubeColorY, cubeColorZ);
        glTranslatef(g_objects[i].xRunning, 0, g_objects[i].zRunning);
        
        glutSolidCube(g_objects[i].size);
        if (isTrip && getRandomFloat(0.0, 1.0) > 0.7) {
            glutWireTeapot(g_objects[i].size);
        }
        g_objects[i].xRunning += g_objects[i].xRate;
        g_objects[i].zRunning += g_objects[i].zRate;
        glPopMatrix();
        
        float howFarBack = getRandomFloat(10.0, 40.0);
        
        if (g_objects[i].zRunning > howFarBack) {
            float positiveOrNegative = getRandomFloat(0, 1);
            g_objects[i].xRate = getRandomFloat(xRateLow, xRateHigh);
            if (positiveOrNegative >= .5) {
                g_objects[i].xRate *= -1;
            }
            
            float randomxStart = getRandomFloat(xStartLow, xStartHigh);
            float positiveOrNegative2 = getRandomFloat(0, 1);
            if (positiveOrNegative2 >= .5) {
                randomxStart *= -1;
            }
            g_objects[i].xRunning = randomxStart;
            g_objects[i].zRunning = 0;
            g_objects[i].zRate = getRandomFloat(zRateLow, zRateHigh);
        }
    }
    
    /* ==
     FFT code
     == */
    if (isTrip) {
        glLineWidth(4);
    } else {
        glLineWidth( 1 );
    }
    // take forward FFT; result in buffer as FFT_SIZE/2 complex values
    rfft( (float *)buffer, g_bufferSize/2, FFT_FORWARD );
    // cast to complex
    complex * cbuf = (complex *)buffer;
    
    xStart = -6.5;
    yStart = 0;
    double inc = 1.0f;
    
    // copy current magnitude spectrum into waterfall memory
    for( int i = 0; i < g_fft_size/2; i++ )
    {
        // copy x coordinate
        g_spectrums[g_wf][i].x = xStart;
        // copy y, depending on scaling
        double freq = cmp_abs(cbuf[i] );
        g_spectrums[g_wf][i].y = yStart + ::pow(freq, .5);
        // increment x
        xStart += inc;
    }
    
//     draw the right things
    g_draw[g_wf] = g_wutrfall;
    if( !g_starting )
        g_draw[(g_wf+g_wf_delay)%g_depth] = true;
    
    glPushMatrix();
    {
        glTranslatef(0.0, 0.0, 0.0);
        glScalef(0.015, 6, 1.0);
        for (int i = 0; i < g_depth; i++) {
            float fval = (float)i / g_depth;
            if (i < g_depth / 4) {
                fval = .25f;
            } else if (i > g_depth / 4 && i < g_depth/2) {
                fval = .5f;
            } else if (i > g_depth/2 && i < g_depth * .75) {
                fval = .75f;
            } else {
                glColor3f(.690, 0.878, 0.902);
            }
            
            fval = (float)i / g_depth;
            
            if (isTrip) {
                glColor3f(getRandomFloat(0.0, 1.0), getRandomFloat(0.0, 1.0), getRandomFloat(0.0, 1.0));
            } else {
                glColor3f(0.529 * fval, 0.808 * fval , 0.980 * fval);
            }
            
            // get the magnitude spectrum of layer
            Pt2D * pt = g_spectrums[(g_wf+i)%g_depth];
            Pt2D * pt2 = pt;
            
            glBegin(GL_LINE_STRIP);
            for (int j = 0; j < g_bufferSize/2; j++, pt++) {
                // i is distance between histories
                float d = (float)i;
                glVertex2f( g_log_positions[j], pt->y + .01*d );
            }
            glEnd();
            
            pt = pt2;
            
            glBegin(GL_LINE_STRIP);
            for (int j = 0; j < g_bufferSize/2; j++, pt++) {
                // i is distance between histories
                float d = (float)i;
                glVertex2f( -g_log_positions[j], pt->y + .01*d );
            }
            glEnd();
        }
    }
    glPopMatrix();
    
    
//     advance index
    g_wf--;
    // mod
    g_wf = (g_wf + g_depth) % g_depth;
    
    
    // flush!
    glFlush( );
    // swap the double buffer
    glutSwapBuffers( );
}

void drawLissajous( SAMPLE * stereobuffer, int len, int channels)
{
    float x, y;
    SAMPLE * buffer;
    
    // 1 or 2 channels only for now
    assert( channels >= 1 && channels <= 2 );

    // mono
    if( channels == 1 )
    {
        buffer = stereobuffer;
        // convert to mono
        for( int m = 0; m < len; m++)
        {
            buffer[m] = stereobuffer[m*2] + stereobuffer[m*2+1];
            buffer[m] /= 2.0f;
        }
    }
    else
    {
        buffer = stereobuffer;
    }
    
    // back to default line width
    glLineWidth( 1.0f );

    // color
    glColor3f( 1.0f, 1.0f, .5f );
    // save current matrix state
    glPushMatrix();
    // translate
    glTranslatef( 1.2f, 0.0f, 0.0f );
    // draw it
    glBegin( GL_LINE_STRIP );
    for( int i = 0; i < len * channels; i += channels )
    {
        x = buffer[i] * g_lissajous_scale;
        if( channels == 1 )
        {
            // delay
            y = (i - g_delay >= 0) ? buffer[i-g_delay] : g_back_buffer[len + i-g_delay];
            y *= g_lissajous_scale;
        }
        else
        {
            y = buffer[i + channels-1] * g_lissajous_scale;
        }

        glVertex3f( x, y, 0.0f );
        // glVertex3f( x, y, sqrt( x*x + y*y ) * -g_lissajous_scale );
    }
    glEnd();
    // restore matrix state
    glPopMatrix();
    
    // hmm...
    if( channels == 1 )
        memcpy( g_back_buffer, buffer, len * sizeof(SAMPLE) );
}

//-----------------------------------------------------------------------------
// Name: map_log_spacing( )
// Desc: ...
//-----------------------------------------------------------------------------
inline double map_log_spacing( double ratio, double power )
{
    // compute location
    return ::pow(ratio, power) * g_fft_size/g_freq_view;
}




//-----------------------------------------------------------------------------
// Name: compute_log_spacing( )x
// Desc: ...
//-----------------------------------------------------------------------------
double compute_log_spacing( int fft_size, double power )
{
    int maxbin = fft_size; // for future in case we want to draw smaller range
    // int minbin = 0; // what about adding this one?

    for(int i = 0; i < fft_size; i++)
    {
        // compute location
        g_log_positions[i] = map_log_spacing( (double)i/fft_size, power );
        // normalize, 1 if maxbin == fft_size
        g_log_positions[i] /= pow((double)maxbin/fft_size, power);
    }

    return 1/::log(fft_size);
}
